@@ -1,6 +1,6 @@
 import ansicolor from 'ansicolor';
 import { LosslessNumber, parse, stringify } from 'lossless-json';
-import Prism, { Grammar } from 'prismjs';
+import Prism, { Grammar, Token } from 'prismjs';
 
 import {
   DataFrame,
@@ -10,21 +10,21 @@ import {
   LogRowModel,
   LogsSortOrder,
   systemDateFormats,
-  textUtil,
 } from '@grafana/data';
+import { t } from '@grafana/i18n';
 import { config } from '@grafana/runtime';
 import { GetFieldLinksFn } from 'app/plugins/panel/logs/types';
 
 import { checkLogsError, checkLogsSampled, escapeUnescapedString, sortLogRows } from '../../utils';
 import { LOG_LINE_BODY_FIELD_NAME } from '../LogDetailsBody';
 import { FieldDef, getAllFields } from '../logParser';
-import { identifyOTelLanguage, getOtelFormattedBody } from '../otel/formats';
+import { identifyOTelLanguage, getOtelAttributesField, OTEL_LOG_LINE_ATTRIBUTES_FIELD_NAME } from '../otel/formats';
 
 import { generateLogGrammar, generateTextMatchGrammar } from './grammar';
 import { LogLineVirtualization } from './virtualization';
 
 const TRUNCATION_DEFAULT_LENGTH = 50000;
-const NEWLINES_REGEX = /(\r\n|\n|\r)/g;
+export const NEWLINES_REGEX = /(\r\n|\n|\r)/g;
 
 export class LogListModel implements LogRowModel {
   collapsed: boolean | undefined = undefined;
@@ -54,11 +54,15 @@ export class LogListModel implements LogRowModel {
   timeUtc: string;
   uid: string;
   uniqueLabels: Labels | undefined;
+  uniqueLabelsExpanded = false;
 
   private _body: string | undefined = undefined;
   private _currentSearch: string | undefined = undefined;
   private _grammar?: Grammar;
   private _highlightedBody: string | undefined = undefined;
+  private _highlightedLogAttributesTokens: Array<string | Token> | undefined = undefined;
+  private _highlightTokens: Array<string | Token> | undefined = undefined;
+  private _escapeUnescapedString = false;
   private _fields: FieldDef[] | undefined = undefined;
   private _getFieldLinks: GetFieldLinksFn | undefined = undefined;
   private _prettifyJSON: boolean;
@@ -108,11 +112,14 @@ export class LogListModel implements LogRowModel {
     this._virtualization = virtualization;
     this._wrapLogMessage = wrapLogMessage;
 
-    let raw = log.raw;
     if (escape && log.hasUnescapedContent) {
-      raw = escapeUnescapedString(raw);
+      this._escapeUnescapedString = true;
+    }
+    this.raw = log.raw;
+
+    if (config.featureToggles.otelLogsFormatting && this.otelLanguage) {
+      this.labels[OTEL_LOG_LINE_ATTRIBUTES_FIELD_NAME] = getOtelAttributesField(this, wrapLogMessage);
     }
-    this.raw = raw;
   }
 
   clone() {
@@ -120,7 +127,8 @@ export class LogListModel implements LogRowModel {
     // Unless this function is required outside of <LogLineDetailsLog />, we create a wrapped clone, so new lines are not stripped.
     clone._wrapLogMessage = true;
     clone._body = undefined;
-    clone._highlightedBody = undefined;
+    clone._highlightTokens = undefined;
+    clone.collapsed = false;
     return clone;
   }
 
@@ -136,7 +144,12 @@ export class LogListModel implements LogRowModel {
           this.raw = reStringified;
         }
       } catch (error) {}
-      const raw = config.featureToggles.otelLogsFormatting && this.otelLanguage ? getOtelFormattedBody(this) : this.raw;
+
+      // always escape for literal \n, \t, \r sequences so "Escape newlines" works for all log types.
+      if (this._escapeUnescapedString) {
+        this.raw = escapeUnescapedString(this.raw);
+      }
+      const raw = this.raw;
       this._body = this.collapsed
         ? raw.substring(0, this._virtualization?.getTruncationLength(null) ?? TRUNCATION_DEFAULT_LENGTH)
         : raw;
@@ -161,14 +174,38 @@ export class LogListModel implements LogRowModel {
   get highlightedBody() {
     if (this._highlightedBody === undefined) {
       // Body is accessed first to trigger the getter code before generateLogGrammar()
-      const sanitizedBody = textUtil.sanitize(this.body);
+      const body = this.body;
       this._grammar = this._grammar ?? generateLogGrammar(this);
       const extraGrammar = generateTextMatchGrammar(this.searchWords, this._currentSearch);
-      this._highlightedBody = Prism.highlight(sanitizedBody, { ...extraGrammar, ...this._grammar }, 'lokiql');
+      this._highlightedBody = Prism.highlight(body, { ...extraGrammar, ...this._grammar }, 'logs');
     }
     return this._highlightedBody;
   }
 
+  get highlightedBodyTokens() {
+    if (this._highlightTokens === undefined) {
+      // Body is accessed first to trigger the getter code before generateLogGrammar()
+      const body = this.body;
+      this._grammar = this._grammar ?? generateLogGrammar(this);
+      const extraGrammar = generateTextMatchGrammar(this.searchWords, this._currentSearch);
+      this._highlightTokens = Prism.tokenize(body, { ...extraGrammar, ...this._grammar });
+    }
+    return this._highlightTokens;
+  }
+
+  get highlightedLogAttributesTokens() {
+    if (this._highlightedLogAttributesTokens === undefined) {
+      const attributes = this.labels[OTEL_LOG_LINE_ATTRIBUTES_FIELD_NAME] ?? '';
+      if (!attributes) {
+        return [];
+      }
+      this._grammar = this._grammar ?? generateLogGrammar(this);
+      const extraGrammar = generateTextMatchGrammar(this.searchWords, this._currentSearch);
+      this._highlightedLogAttributesTokens = Prism.tokenize(attributes, { ...extraGrammar, ...this._grammar });
+    }
+    return this._highlightedLogAttributesTokens;
+  }
+
   get isJSON() {
     return this._json;
   }
@@ -223,21 +260,22 @@ export class LogListModel implements LogRowModel {
     if (this.collapsed === undefined || collapsed === undefined) {
       this.collapsed = collapsed;
       this._body = undefined;
-      this._highlightedBody = undefined;
+      this._highlightTokens = undefined;
     }
   }
 
   setCollapsedState(collapsed: boolean) {
     if (this.collapsed !== collapsed) {
       this._body = undefined;
-      this._highlightedBody = undefined;
+      this._highlightTokens = undefined;
     }
     this.collapsed = collapsed;
   }
 
   setCurrentSearch(search: string | undefined) {
     this._currentSearch = search;
-    this._highlightedBody = undefined;
+    this._highlightTokens = undefined;
+    this._highlightedLogAttributesTokens = undefined;
   }
 }
 
@@ -315,3 +353,20 @@ function countNewLines(log: string, limit = Infinity) {
   }
   return count;
 }
+
+export function getLevelsFromLogs(logs: LogListModel[]) {
+  const levels = new Set<LogLevel>();
+  for (const log of logs) {
+    levels.add(log.logLevel);
+  }
+  return Array.from(levels).filter((level) => level != null);
+}
+
+export function getNormalizedFieldName(field: string) {
+  if (field === LOG_LINE_BODY_FIELD_NAME) {
+    return t('logs.log-line-details.log-line-field', 'Log line');
+  } else if (field === OTEL_LOG_LINE_ATTRIBUTES_FIELD_NAME) {
+    return t('logs.log-line-details.log-attributes-field', 'OTel attributes');
+  }
+  return field;
+}
